/*****************************************************************************/
/*                                                                           */
/* Title: Hidden Markov Model training (Baum-Welch)                          */
/* File: "hmm_train.c"                                                       */
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 5/14/96                                                    */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "matrix.h"
#include "file_io.h"
#include "gen_math.h"
#include "hmm_util.h"
#ifdef MacOS
#include <console.h>
#endif
#include "arg.h"

/***************************** Prototypes ************************************/

HMM *initialize_hmm(HMM *model, int states, int observables, 
		    int bakis_flag, int bakis_connections);

/**************************** Main Program ***********************************/

#define SCALE 1

int main(int argc, char **argv)
{
  FILE *fp;
  char *in_file = NULL, *out_file = NULL, format[STR_REG];
  char *hmm_file = NULL;
  MULTIPLE m;
  HMM *old_model, *new_model;
  int n_states = 5, bakis_flag = NO_BAKIS, bakis_connections = 2;
  int iterations = 100;
  int precision = 15, seed = 1, i, j;
  int detailed_print = 1;
  int convergence_flag;
  int length = 5;
  double convergence = .000001, old_prob = 0., new_prob = 0.;
  double *diff;

#ifdef MacOS
  argc = ccommand(&argv);
#endif

  arg_parse(argc, argv,
	    "", "Usage:",
	    "-input %S", &in_file,  "set input file",
	    "-output %S", &out_file, "set output hmm file",
	    "-initial [%S]", &hmm_file, 
            "read in initial hmm file (-structure -states ignored) [none]",
	    "-structure [%d]", &bakis_flag, 
	    "[0 = Fully connected, 1 = Bakis, 2 = Looping Bakis] [%d]",
	    bakis_flag, 
	    "-forward [%d]", &bakis_connections,
	    "set number offorward connections (only for structure 1,2) [%d]",
	    bakis_connections,
	    "-convergence [%F]", &convergence,    
	    "set convergence criterion [%lf]", convergence,
	    "-length [%d]", &length, 
	    "number of iterations to check for convergence criterion [%d]",
	    length,
	    "-states [%d]", &n_states,
	    "set number of states [%d]", n_states,
	    "-iterations [%d]", &iterations,
	    "set max. number of iterations [%d]",
	    iterations,
	    "-precision [%d]", &precision, 
	    "set output precision [%d]", precision,
	    "-seed [%d]", &seed, "random seed [%d]", seed,
	    "-print [%d]", &detailed_print, "print per iteration report [%d]",
	    detailed_print,
	    NULL);

  if(in_file == NULL || out_file == NULL) {
    fprintf(stderr, "No input and/or output file specified.\n");
    exit(-1);
  }

  srand(seed);

  read_data_file(&m, in_file);

  if(m.n_total <= 0) {
    fprintf(stderr, "Warning: symbol file contains no data.\n");
    fprintf(stderr, "Defaulting to single-state, uniform HMM.\n");
    new_model = allocate_hmm(new_model, 1, m.n_observables);

    for(i = 0; i < m.n_observables; i++)
      new_model->B[i][0] = 1./((double) m.n_observables);

    new_model->B = normalize_matrix(new_model->B, m.n_observables, 1, COLUMN);
    new_model->A[0][0] = 1.;
    new_model->Pi[0] = 1.;
   
    fp = get_fp(out_file, stdout, fp);
    display_hmm(fp, new_model, precision);
    fclose(fp);

    exit(1);
  }

  if(hmm_file == NULL) {
    old_model = allocate_hmm(old_model, n_states, m.n_observables);
    old_model = initialize_hmm(old_model, n_states, m.n_observables, 
			     bakis_flag, bakis_connections);
    new_model = allocate_hmm(new_model, n_states, m.n_observables);
  } else {
    old_model = read_hmm(old_model, hmm_file);
    new_model = allocate_hmm(new_model, old_model->n_states,
			     old_model->n_observables);
  }

  length--;
  diff = allocate_vector(diff, iterations);

  fprintf(stdout, "Total data: %d\n", m.n_total);

  for(i = 0; i < iterations; i++) {
    new_model = baum_welch(new_model, old_model, m.obs, m.n_sequence, 
			   0., bakis_flag, SCALE);

    new_prob = exp10(new_model->prob/((double) m.n_total));

    diff[i] = fabs(new_prob - old_prob)/new_prob;

    if(detailed_print == 1) {
      fprintf(stdout, "Iteration # %d : %.7f %.7f ( %.15f )\n", 
	      i, new_model->prob, new_prob, diff[i]);
    }

    convergence_flag = 1;
    for(j = max_int(i - length, 0); j <= i; j++)
      convergence_flag &= (diff[j] < convergence);
    
    if(convergence_flag > 0) {
      if(detailed_print == 0) {
	fprintf(stdout, "Iteration # %d : %.7f %.7f ( %.15f )\n", 
		i, new_model->prob, new_prob, diff[i]);
      }
      break;
    }

    if(new_prob < old_prob)
      fprintf(stderr, "Alert: Probability decreasing at iteration #%d\n", i);

    old_prob = new_prob;

    old_model = copy_hmm(old_model, new_model);
  }

  fp = get_fp(out_file, stdout, fp);
  display_hmm(fp, new_model, precision);
  fclose(fp);
}

/***************************** Subroutines ***********************************/

HMM *initialize_hmm(HMM *model, int states, int observables, 
		    int bakis_flag, int bakis_connections)
{
  if(bakis_flag == BAKIS) {
    model->A = random_bakis(model->A, states, states, 
			    bakis_connections);
    model->B = random_B(model->B, observables, states);
    model->Pi = zero_vector(model->Pi, states);
    model->Pi[0] = 1.;
  } else if(bakis_flag == CONTINUED_BAKIS) {
    model->A = random_continued_bakis(model->A, states, states, 
				      bakis_connections);
    model->B = random_B(model->B, observables, states);
    model->Pi = zero_vector(model->Pi, states);
    model->Pi[0] = 1.;
  } else {
    model->A = random_A(model->A, states, states);
    model->B = random_B(model->B, observables, states);
    model->Pi = random_Pi(model->Pi, states);
  }

  return(model);
}

/*****************************************************************************/
