/*****************************************************************************/
/*                                                                           */
/* Title: Hidden Markov Model Simulator Functions                            */
/* File: "hmm_util.c"                                                        */
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 11/8/2002                                                  */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "matrix.h"
#include "file_io.h"
#include "gen_math.h"
#include "hmm_util.h"

/***************************** Subroutines ***********************************/

HMM **read_multiple_hmms(HMM **model, char *in_file, int n)
{
  int i;
  char str[256];

  model = (HMM **) malloc(n * sizeof(HMM *));
  is_null((int *) model);

  for(i = 0; i < n; i++) {
    sprintf(str, "%s.%d", in_file, i);
    model[i] = read_hmm(model[i], str);
  }

  return(model);
}

/*****************************************************************************/

HMM *read_hmm(HMM *model, char *in_file)
{
  FILE *fp;
  int n_states, n_observables;

  fp = fopen(in_file, "r");

  if(fp == NULL) {
    fprintf(stderr, "File %s does not exist\n", in_file);
    return((HMM *) fp);
  }

  fscanf(fp, "%*s %d", &n_states);
  fscanf(fp, "%*s %d", &n_observables);

  model = allocate_hmm(model, n_states, n_observables);

  strcpy(&model->name[0], in_file); /* only used for hmm_mult_eval */

  /* Read in A, B, Pi matrices */

  fscanf(fp, "%*s");
  model->A = read_matrix(fp, model->A, model->n_states, 0);

  fscanf(fp, "%*s");
  model->B = read_matrix(fp, model->B, model->n_observables, 
			    model->n_states);

  fscanf(fp, "%*s");
  model->Pi = read_vector(fp, model->Pi, model->n_states);

  fclose(fp);

  return(model);
}

/*****************************************************************************/

void display_sequence(FILE *fp, SEQUENCE *obs)
{
  int i;

  for(i = 0; i < obs->T; i++)
    fprintf(fp, "%d, ", obs->sequence[i]);

  fprintf(fp, "\n");
}

/*****************************************************************************/

void display_hmm(FILE *fp, HMM *model, double precision)
{
  fprintf(fp, "STATES %d\nOBSERVABLES %d\n\n",
	  model->n_states, model->n_observables);

  fprintf(fp, "A-MATRIX\n\n");
  display_matrix(fp, model->A, model->n_states, 0, precision);

  fprintf(fp, "\nB-MATRIX:\n\n");
  display_matrix(fp, model->B, model->n_observables, model->n_states, 
		 precision);

  fprintf(fp, "\nPI-VECTOR\n\n");
  display_vector(fp, model->Pi, model->n_states, 0, precision);
}

/*****************************************************************************/

void file_hmm(char *out_file, HMM *model, double precision)
{
  FILE *fp;

  fp = fopen(out_file, "w");
  display_hmm(fp, model, precision);
  fclose(fp);
}

/*****************************************************************************/

HMM *allocate_hmm(HMM *model, int n_states, int n_observables)
{
  int i;

  model = (HMM *) malloc(sizeof(HMM));
  is_null((int *) model);

  model->n_states = n_states;
  model->n_observables = n_observables;

  model->A = allocate_matrix(model->A, model->n_states, 0);
  model->B = allocate_matrix(model->B, model->n_observables, model->n_states);
  model->Pi = allocate_vector(model->Pi, n_states);

  model->prob = 0.;

  return(model);
}

/*****************************************************************************/

SEQUENCE *allocate_sequence(SEQUENCE *obs, int n)
{
  obs = (SEQUENCE *) malloc(sizeof(SEQUENCE));
  is_null((int *) obs);

  obs->sequence = (int *) malloc(n * sizeof(int));
  is_null((int *) obs->sequence);

  obs->T = n;

  return(obs);
}

/*****************************************************************************/

FB *forward_backward(HMM *model, SEQUENCE *obs, FB *ab, int scale)
{
  ab = forward_calculation(model, obs, ab, scale);
  ab = backward_calculation(model, obs, ab);

  return(ab);
}

/*****************************************************************************/

FB *forward_calculation(HMM *model, SEQUENCE *obs, FB *ab, int scale)
{
  int i, j, t;

  /* 1) Initialization */

  for(i = 0; i < model->n_states; i++) {
    ab->alpha[i][0] = model->Pi[i] * model->B[obs->sequence[0]][i];
  }

  ab = scale_alpha(model, ab, 0, scale);

  /* 2) Induction */

  for(t = 1; t < obs->T; t++) {
    for(j = 0; j < model->n_states; j++) {
      ab->alpha[j][t] = 0.;
      for(i = 0; i < model->n_states; i++)
	ab->alpha[j][t] += (ab->alpha[i][t-1] * model->A[i][j]);
      ab->alpha[j][t] *= model->B[obs->sequence[t]][j];
    }
    ab = scale_alpha(model, ab, t, scale);
  }
  
  return(ab);
}

/*****************************************************************************/

FB *backward_calculation(HMM *model, SEQUENCE *obs, FB *ab)
{
  int i, j, t;

  /* 1) Initialization */

  for(i = 0; i < model->n_states; i++)
    ab->beta[i][obs->T - 1] = 1.;

  ab = scale_beta(model, ab, obs->T - 1);

  /* 2) Induction */

  for(t = obs->T - 2; t >= 0 ; t--) {
    for(i = 0; i < model->n_states; i++) {
      ab->beta[i][t] = 0.;
      for(j = 0; j < model->n_states; j++)
	ab->beta[i][t] += 
	  (model->A[i][j] * model->B[obs->sequence[t+1]][j] *
	   ab->beta[j][t+1]);
    }
    ab = scale_beta(model, ab, t);
  }
  
  return(ab);
}

/*****************************************************************************/

FB *allocate_ab(HMM *model, SEQUENCE *obs, FB *ab)
{
  ab = (FB *) malloc(sizeof(FB));

  is_null((int *) ab);

  ab->alpha = allocate_matrix(ab->alpha, model->n_states, obs->T);
  ab->beta  = allocate_matrix(ab->beta,  model->n_states, obs->T);

  ab->c = allocate_vector(ab->c, obs->T);

  return(ab);
}

/*****************************************************************************
 *
 * Function: prob_obs
 * Description: returns log[P2(O | lambda)]. Alpha's, beta's and scaling
 *              coefficients c's are already assumed calculated.
 *
 *****************************************************************************/

double prob_obs(HMM *model, SEQUENCE *obs, FB *ab, int scale)
{
  int i;
  double tmp = 0.;

  if(scale > 0) {
    for(i = 0; i < obs->T; i++)
      tmp += log10(ab->c[i]);

    tmp = (-tmp);
  } else {
    for(i = 0; i < model->n_states; i++)
      tmp += ab->alpha[i][obs->T - 1];

    tmp = log10(tmp);
  }

  return(tmp);
}

/*****************************************************************************/

FB *scale_alpha(HMM *model, FB *ab, int t, int scale)
{
  int i;

  ab->c[t] = 0.;

  if(scale > 0) {
    for(i = 0; i < model->n_states; i++)
      ab->c[t] += ab->alpha[i][t];
    ab->c[t] = 1./ab->c[t]; 
    for(i = 0; i < model->n_states; i++)
      ab->alpha[i][t] *= ab->c[t];
  } else
    ab->c[t] = 1.;

  return(ab);
}

/*****************************************************************************/

FB *scale_beta(HMM *model, FB *ab, int t)
{
  int i;

  for(i = 0; i < model->n_states; i++)
    ab->beta[i][t] *= ab->c[t];

  return(ab);
}

/*****************************************************************************/

HMM *baum_welch_one(HMM *new_model, HMM *old_model, 
		    SEQUENCE *obs, double smooth_threshold, int scale)
{
  new_model = baum_welch(new_model, old_model, &obs, 1, smooth_threshold, 
			 NO_BAKIS, scale);
			 
  return(new_model);
}

/*****************************************************************************/

HMM *baum_welch(HMM *new_model, HMM *old_model, 
		SEQUENCE **obs, int n_sequence, double smooth_threshold,
		int bakis_model, int scale)
{
  FB **ab;
  double *denominator, *inv_p, tmp;
  int k;

  ab = (FB **) malloc(n_sequence * sizeof(FB *));
  is_null((int *) ab);

  for(k = 0; k < n_sequence; k++)
    ab[k] = allocate_ab(old_model, obs[k], ab[k]);
  
  denominator = allocate_vector(denominator, old_model->n_states);
  inv_p = allocate_vector(inv_p, n_sequence);

  for(k = 0; k < n_sequence; k++) {
    ab[k] = forward_backward(old_model, obs[k], ab[k], scale);
    /*
      fprintf(stdout, "\nalphas:\n");
      display_matrix(stdout, ab[k]->alpha, old_model->n_states, obs[k]->T, 15);
      fprintf(stdout, "\nbetas:\n");
      display_matrix(stdout, ab[k]->beta, old_model->n_states, obs[k]->T, 15);
      fprintf(stdout, "\ncs:\n");
      display_vector(stdout, ab[k]->c, obs[k]->T, 0, 15);
      */
    inv_p[k] = prob_obs(old_model, obs[k], ab[k], scale);
  }

  inv_p = adjust_inv_p(inv_p, n_sequence, scale);

  /*
    fprintf(stdout, "\ninv_p:\n");
    display_vector(stdout, inv_p, n_sequence, 0, 15);
  */

  denominator = bw_denominator(denominator, old_model, obs, n_sequence, ab, 
			       inv_p);

  /*
    fprintf(stdout, "\nA denominator:\n");
    display_vector(stdout, denominator, old_model->n_states, 0, 15);
    */

  new_model = bw_A(new_model, old_model, obs, n_sequence, ab,
		   denominator, inv_p);
  /*
    fprintf(stdout, "\nA matrix:\n");
    display_matrix(stdout, new_model->A, old_model->n_states, 
    old_model->n_states, 15);
    */

  new_model = bw_B(new_model, old_model, obs, n_sequence, ab,
		   denominator, inv_p);

  /*
    fprintf(stdout, "\nB denominator:\n");
    display_vector(stdout, denominator, old_model->n_states, 0, 15);

    fprintf(stdout, "\nB matrix:\n");
    display_matrix(stdout, new_model->B, old_model->n_observables, 
    old_model->n_states, 15);
    */

  new_model = bw_Pi(new_model, ab, bakis_model);

  new_model = bw_smooth(new_model, smooth_threshold, 1);

  new_model->prob = 0.;
  for(k = 0; k < n_sequence; k++)
    new_model->prob += prob_obs(new_model, obs[k], ab[k], scale);

  for(k = 0; k < n_sequence; k++)
    free_ab(new_model, obs[k], ab[k]);

  free(ab);
  free(denominator);
  free(inv_p);

  return(new_model);
}

/*****************************************************************************/

double probability(HMM *model, SEQUENCE **obs, int n_sequence, int scale,
		   double *probs)
{
  int i,j,k;
  FB **ab;

  ab = (FB **) malloc(n_sequence * sizeof(FB *));
  is_null((int *) ab);

  model->prob = 0.;

  for(k = 0; k < n_sequence; k++) {
    ab[k] = allocate_ab(model, obs[k], ab[k]);
    ab[k] = forward_backward(model, obs[k], ab[k], scale);
    probs[k] = prob_obs(model, obs[k], ab[k], scale);
    model->prob += probs[k];

    /* TEMPORARY */

    /*
    compute_matrix_log10(ab[k]->alpha, model->n_states, obs[k]->T);
    display_matrix(stdout, ab[k]->alpha, model->n_states, obs[k]->T, 10);
    fprintf(stdout, "\n");

    compute_matrix_log10(ab[k]->beta, model->n_states, obs[k]->T);
    display_matrix(stdout, ab[k]->beta,  model->n_states, obs[k]->T, 10);
    fprintf(stdout, "\n");

    compute_vector_log10(ab[k]->c, obs[k]->T);
    display_vector(stdout, ab[k]->c,  obs[k]->T, 0, 10);
    fprintf(stdout, "\n");
    */

    free_ab(model, obs[k], ab[k]);
  }

  free(ab);
  return(model->prob);
}

/*****************************************************************************/

double *adjust_inv_p(double *inv_p, int n, int scale)
{
  int i;
  double min_ceil;

  min_ceil = ceil(min_vec(inv_p, n));

  for(i = 0; i < n; i++) {
    if(scale > 0)
      inv_p[i] = 1.;
    else
      inv_p[i] = 1./pow(10., inv_p[i] - min_ceil);
  }

  return(inv_p);
}

/*****************************************************************************/

double *bw_denominator(double *denominator, HMM *old_model,
		       SEQUENCE **obs, int n_sequence,
		       FB **ab, double *inv_p)
{
  int j, k, t;
  double tmp;
  
  for(j = 0; j < old_model->n_states; j++) {
    denominator[j] = 0.;
    for(k = 0; k < n_sequence; k++) {
      tmp = 0.;
      for(t = 0; t < obs[k]->T - 1; t++)
	tmp += (ab[k]->alpha[j][t] * ab[k]->beta[j][t])/ab[k]->c[t];
      tmp *= inv_p[k];
      denominator[j] += tmp;
    }
  }

  return(denominator);
}

/*****************************************************************************/

double *bw_denominator_adjust(double *denominator, HMM *old_model,
			      SEQUENCE **obs, int n_sequence,
			      FB **ab, double *inv_p)
{
  int j, k, t;

  for(j = 0; j < old_model->n_states; j++)
    for(k = 0; k < n_sequence; k++) {
      t = obs[k]->T - 1;
      denominator[j] += (ab[k]->alpha[j][t] * ab[k]->beta[j][t])/ab[k]->c[t]
	* inv_p[k];
    }

  return(denominator);
}

/*****************************************************************************/

HMM *bw_A(HMM *new_model, HMM *old_model, 
	  SEQUENCE **obs, int n_sequence, FB **ab,
	  double *denominator, double *inv_p)
{
  int i, j, k, t;
  double tmp;

  new_model->A = zero_matrix(new_model->A, new_model->n_states,
			     new_model->n_states);

  for(i = 0; i < old_model->n_states; i++)
    for(j = 0; j < old_model->n_states; j++) {
      if(old_model->A[i][j] > SMALL*10.) {
	for(k = 0; k < n_sequence; k++) { 
	  tmp = 0.;
	  for(t = 0; t < obs[k]->T - 1; t++) {
	    tmp += (ab[k]->alpha[i][t] *
		     old_model->A[i][j] *
		     old_model->B[obs[k]->sequence[t+1]][j] *
		     ab[k]->beta[j][t+1]);
	  }
	  tmp *= inv_p[k];
	  new_model->A[i][j] += tmp;
	}
	if(!is_zero(denominator[i])) {           /* degenerative case:        */
	  new_model->A[i][j] /= denominator[i]; /* all sequences of length 1 */
	} else {                                /* hmm states = 1            */
	  new_model->A[i][j] = 1.;
	}
      } else {
	new_model->A[i][j] = 0.;
      }
    }

  return(new_model);
}

/*****************************************************************************/

HMM *bw_B(HMM *new_model, HMM *old_model, 
	  SEQUENCE **obs, int n_sequence, FB **ab,
	  double *denominator, double *inv_p)
{
  int i, j, k, t;
  double *b_tmp, **gamma_table;

  new_model->B = zero_matrix(new_model->B, new_model->n_observables,
			     new_model->n_states);

  b_tmp = allocate_vector(b_tmp, old_model->n_observables);

  denominator = bw_denominator_adjust(denominator, old_model, obs,
				      n_sequence, ab, inv_p);

  for(j = 0; j < old_model->n_states; j++) {

    for(k = 0; k < n_sequence; k++) {
      for(i = 0; i < old_model->n_observables; i++)
	b_tmp[i] = 0.;
      
      for(t = 0; t < obs[k]->T; t++)
	b_tmp[obs[k]->sequence[t]] += 
	  (ab[k]->alpha[j][t] * ab[k]->beta[j][t])/ab[k]->c[t];

      for(i = 0; i < old_model->n_observables; i++) {
	b_tmp[i] *= inv_p[k];
	new_model->B[i][j] += b_tmp[i];
      }
    }
    
    for(i = 0; i < old_model->n_observables; i++)
      new_model->B[i][j] /= denominator[j];
  }

  free(b_tmp);
  return(new_model);
}

/*****************************************************************************/

HMM *bw_Pi(HMM *new_model, FB **ab, int flag)
{
  int i;

  if(flag == BAKIS) {
    for(i = 1; i < new_model->n_states; i++)
      new_model->Pi[i] = 0.;

    new_model->Pi[0] = 1.;

  } else {
    double sum = 0.;

    for(i = 0; i < new_model->n_states; i++) {
      new_model->Pi[i] =  ab[0]->alpha[i][0] * ab[0]->beta[i][0];
      sum += new_model->Pi[i];
    }

    for(i = 0; i < new_model->n_states; i++)
      new_model->Pi[i] /= sum;
  }

  return(new_model);
}

/*****************************************************************************/

HMM *bw_smooth(HMM *model, double threshold, int norm)
{
  model->A = threshold_matrix(model->A, model->n_states, model->n_states, 
			      threshold);
  model->B = threshold_matrix(model->B, model->n_observables, model->n_states, 
			      threshold);
  model->Pi = threshold_vector(model->Pi, model->n_states, threshold);

  if(norm > 0) {
    model->A = normalize_matrix(model->A, model->n_states, 
				model->n_states, ROW);
    model->B = normalize_matrix(model->B, model->n_observables, 
				model->n_states, COLUMN);
    model->Pi = normalize_vector(model->Pi, model->n_states);
  }
  return(model);
}

/*****************************************************************************/

double **normalize_matrix(double **matrix, int n, int m, int row_or_column)
{
  int i, j;
  int iterate_1, iterate_2;
  double tmp;

  iterate_1 = (row_or_column == ROW) ? n : m;
  iterate_2 = (row_or_column == COLUMN) ? n : m;

  for(i = 0; i < iterate_1; i++) {
    tmp = 0.;
    for(j = 0; j < iterate_2; j++)
        tmp += (row_or_column == ROW) ? matrix[i][j] : matrix[j][i];
    for(j = 0; j < iterate_2; j++) {
      if(row_or_column == ROW)
        matrix[i][j] /= tmp;
      else
        matrix[j][i] /= tmp;
    }
  }

  return(matrix);
}

/*****************************************************************************/

double *random_Pi(double *vector, int n)
{
  vector = random_vector(vector, n, 1., 2.);
  vector = normalize_vector(vector, n);

  return(vector);
}

/*****************************************************************************/

double **random_A(double **matrix, int n, int m)
{
  matrix = random_matrix(matrix, n, m, 1., 2.);
  matrix = normalize_matrix(matrix, n, m, ROW);
  
  return(matrix);
}

/*****************************************************************************/

double **random_B(double **matrix, int n, int m)
{
  matrix = random_matrix(matrix, n, m, 1., 2.);
  matrix = normalize_matrix(matrix, n, m, COLUMN);

  return(matrix);
}

/*****************************************************************************/

double **random_bakis(double **matrix, int n, int m, int connections)
{
  int i, j;

  matrix = zero_matrix(matrix, n, m);

  for(i = 0; i < n; i++)
    for(j = i; j < i + connections; j++) {
      if(j < n)
	matrix[i][j] = (double) random_integer(INT_RAND);
    }

  matrix = normalize_matrix(matrix, n, m, ROW);

  return(matrix);
}

/*****************************************************************************/

double **random_continued_bakis(double **matrix, int n, int m, 
				int connections)
{
  int i, j;

  matrix = zero_matrix(matrix, n, m);

  for(i = 0; i < n; i++)
    for(j = i; j < i + connections; j++) {
      if(j < n)
	matrix[i][j] = (double) random_integer(INT_RAND);
      else
	matrix[i][j-n] = (double) random_integer(INT_RAND);
    }

  matrix = normalize_matrix(matrix, n, m, ROW);

  return(matrix);
}

/*****************************************************************************/

double *normalize_vector(double *vector, int n)
{
  int i;
  double sum = 0.;

  for(i = 0; i < n; i++)
    sum += vector[i];

  for(i = 0; i < n; i++)
    vector[i] /= sum;

  return(vector);
}

/*****************************************************************************/

HMM *copy_hmm(HMM *old_model, HMM *new_model)
{
  old_model->A = copy_matrix(new_model->A, old_model->A, 
			     new_model->n_states, new_model->n_states);
  old_model->B = copy_matrix(new_model->B, old_model->B,
			     new_model->n_observables, new_model->n_states);
  
  old_model->Pi = copy_vector(new_model->Pi, old_model->Pi,
			      new_model->n_states);

  old_model->prob = new_model->prob;

  return(old_model);
}

/*****************************************************************************/

FB *free_ab(HMM *model, SEQUENCE *obs, FB *ab)
{
  ab->alpha = free_matrix(ab->alpha, model->n_states);
  ab->beta  = free_matrix(ab->beta,  model->n_states);

  free(ab->c);
  free(ab);

  return(ab);
}

/*****************************************************************************/

HMM *free_hmm(HMM *model, int n_states, int n_observables)
{
  model->A = free_matrix(model->A, model->n_states);
  model->B = free_matrix(model->B, model->n_observables);

  free(model->Pi);
  free(model);

  return(model);
}

/*****************************************************************************/

SEQUENCE *free_sequence(SEQUENCE *obs, int n)
{
  free(obs->sequence);
  free(obs);

  return(obs);
}

/*****************************************************************************/

MULTIPLE *read_data_file(MULTIPLE *m, char *file_name)
{
  FILE *fp;
  int i, length;

  fp = fopen(file_name, "r");
  if(fp == NULL) {
    fprintf(stderr, "No file of that name.\n");
    exit(1);
  }

  fscanf(fp, "%d", &m->n_sequence);
  fscanf(fp, "%d", &m->n_observables);

  m->obs = (SEQUENCE **) malloc(m->n_sequence * sizeof(SEQUENCE *));
  is_null((int *) m->obs);

  for(i = 0; i < m->n_sequence; i++)
    m->obs[i] = read_one_sequence(fp, m->obs[i]);
    
  fclose(fp);

  m->n_total = 0;
  for(i = 0; i < m->n_sequence; i++)
    m->n_total += m->obs[i]->T;

  return(m);
}

/*****************************************************************************/

SEQUENCE *read_one_sequence(FILE *fp, SEQUENCE *obs)
{
  int i;
  
  obs = (SEQUENCE *) malloc(sizeof(SEQUENCE));
  is_null((int *) obs);

  fscanf(fp, "%d", &obs->T);
  fscanf(fp, "%s", &obs->name[0]);

  obs->sequence = (int *) malloc(obs->T * sizeof(int));
  is_null((int *) obs->sequence);

  for(i = 0; i < obs->T; i++)
    fscanf(fp, "%d", &obs->sequence[i]);

  return(obs);
}

/*****************************************************************************/

SEQUENCE *generate_sequence(HMM *model, SEQUENCE *obs, int seed)
{
  int i, ns, no, current_state;
  double **a, **bt, **b, *pi;

  srand(seed);

  ns = model->n_states;
  no = model->n_observables;

  a = allocate_matrix(a, ns, ns + 1);
  bt = allocate_matrix(b, ns, no);
  b = allocate_matrix(b, ns, no + 1);
  pi = allocate_vector(pi, ns + 1);
  obs->sequence = int_allocate_vector(obs->sequence, obs->T);

  bt = transpose_matrix(model->B, no, ns, bt);

  for(i = 0; i < ns; i++) {
    transform_vector(a[i], model->A[i], ns);
    transform_vector(b[i], bt[i], no);
  }
    
  transform_vector(pi, model->Pi, ns);

  current_state = which_one(pi, ns);

  for(i = 0; i < obs->T; i++) {
    obs->sequence[i] = which_one(b[current_state], no);
    current_state = which_one(a[current_state], ns);
  }

  /*
  free_matrix(a, ns);
  free_matrix(b, ns);
  free_matrix(bt, no);
  free_vector(pi);
  */
  return(obs);
}

/*****************************************************************************/

int which_one(double *vec, int n)
{
  IntervalHMM tst;
  double x;

  tst.low = 0;
  tst.high = n;
  x = random_real2(0., 1.);

  /* display_vector(stderr, vec, n+1,0,3); */

  binary_search_hmm(vec, &tst, x);
  
  if(!(x >= vec[tst.low] && x <= vec[tst.high])) {
    fprintf(stderr, "Error (%d, %d), (%lf, %lf, %lf)\n",
	    tst.low, tst.high,
	    vec[tst.low], x, vec[tst.high]);
  }

  return(tst.low);
}

/*****************************************************************************/

void binary_search_hmm(double *vec, IntervalHMM *n, double x)
{
  int nm, nl, nh;

  nl = n->low;
  nh = n->high;

  if(nl == nh || abs(nl - nh)==1)
    return;
  else {
    nm = (nl + nh)/2;
    if(vec[nm] < x) {
      n->low = nm;
      binary_search_hmm(vec, n, x);
    } else {
      n->high = nm;
      binary_search_hmm(vec, n, x);
    }
  }
}

/*****************************************************************************/

double *transform_vector(double *out, double *in, int n)
{
  int i;

  out[0] = 0.;

  for(i = 1; i <= n; i++) {
    out[i] = in[i-1] + out[i-1];
  }

  return(out);
}

/*****************************************************************************/

double viterbi(SEQUENCE *states, HMM *model, SEQUENCE *obs)
{
  int i, j, t, **psi;
  double **phi, tmp[100], logprob;


  states->T = obs->T;
  phi = allocate_matrix(phi, obs->T, model->n_states);
  psi = int_allocate_matrix(psi, obs->T, model->n_states);

  for(i = 0; i < model->n_states; i++)
    phi[0][i] = log10(model->Pi[i]) + 
      log10(model->B[obs->sequence[0]][i]);

  for(t = 1; t < obs->T; t++) {
    for(j = 0; j < model->n_states; j++) {
      for(i = 0; i < model->n_states; i++)
	tmp[i] = phi[t-1][i] + log10(model->A[i][j]);
      phi[t][j] = max_vec(&tmp[0], model->n_states) +
	log10(model->B[obs->sequence[t]][j]);
      psi[t][j] = ind_max_vec(&tmp[0], model->n_states);
    }
  }

  states->sequence[obs->T-1] = ind_max_vec(&phi[obs->T-1][0], model->n_states);

  for(t = obs->T-2; t >= 0; t--) {
    states->sequence[t] = psi[t+1][states->sequence[t+1]];
  }

  int_free_matrix(psi, obs->T);
  free_matrix(phi, obs->T);

  logprob = max_vec(&phi[obs->T-1][0], model->n_states);

  return(logprob);
}

/*****************************************************************************/
