/*****************************************************************************/
/*                                                                           */
/* Title: Hidden Markov Model utility functions (header file)                */
/* File: "hmm_util.h"                                                        */
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 11/8/2002                                                  */
/*                                                                           */
/*****************************************************************************/

/******************************* Macros **************************************/

#define NO_BAKIS        0
#define BAKIS           1
#define CONTINUED_BAKIS 2

#define ROW    1
#define COLUMN 2

#define INT_RAND RAND_MAX

#define FILES_MAX 60

/***************************** Structures ************************************/

typedef struct interval_hmm_struct {
  int low;
  int high;
} IntervalHMM;

typedef struct sequence
{
  int *sequence;
  int T;
  char name[STR_REG];
} SEQUENCE;

typedef struct multiple
{
  SEQUENCE **obs;
  int n_total;
  int n_observables;
  int n_sequence;
} MULTIPLE;

typedef struct fb
{
  double **alpha;
  double **beta;
  double *c;
} FB;

typedef struct hmm
{
  char name[256]; /* only used in hmm_mult_eval */
  int n_states;
  int n_observables;

  double **A;
  double **B;
  double *Pi;
  double prob;
} HMM;

typedef char  str[256];

/***************************** Prototypes ************************************/

/* HMM utility functions */

extern void file_hmm(char *out_file, HMM *model, double precision);
extern void display_hmm(FILE *fp, HMM *model, double precision);
extern void display_sequence(FILE *fp, SEQUENCE *obs);

extern HMM **read_multiple_hmms(HMM **model, char *in_file, int n);
extern HMM *read_hmm(HMM *model, char *in_file);
extern HMM *copy_hmm(HMM *old_model, HMM *new_model);

/* Matrix manipulations */

extern double **normalize_matrix(double **matrix, int n, int m, 
				 int row_or_column);

/* Vector manipulations */

extern double *normalize_vector(double *vector, int n);

/* Memory allocations */

extern FB *free_ab(HMM *model, SEQUENCE *obs, FB *ab);
extern HMM *free_hmm(HMM *model, int n_states, int n_observables);
extern SEQUENCE *free_sequence(SEQUENCE *obs, int n);

extern SEQUENCE *allocate_sequence(SEQUENCE *obs, int n);
extern HMM *allocate_hmm(HMM *model, int n_states, int n_observables);
extern FB *allocate_ab(HMM *model, SEQUENCE *obs, FB *ab);

/* Forward-Backward algorithm */

extern FB *forward_backward(HMM *model, SEQUENCE *obs, FB *ab, int scale);
extern FB *forward_calculation(HMM *model, SEQUENCE *obs, FB *ab, int scale);
extern FB *backward_calculation(HMM *model, SEQUENCE *obs, FB *ab);

extern FB *scale_alpha(HMM *model, FB *ab, int t, int scale);
extern FB *scale_beta(HMM *model, FB *ab, int t);

/* Evaluation (probability) */

extern double prob_obs(HMM *model, SEQUENCE *obs, FB *ab, int scale);

/* Baum-Welch Functions */

extern double *bw_denominator(double *denominator, HMM *old_model,
			      SEQUENCE **obs, int n_sequence,
			      FB **ab, double *inv_p);

extern double *bw_denominator_adjust(double *denominator, HMM *old_model,
			      SEQUENCE **obs, int n_sequence,
			      FB **ab, double *inv_p);

extern double *adjust_inv_p(double *inv_p, int n, int scale);

extern HMM *baum_welch_one(HMM *new_model, HMM *old_model, 
			   SEQUENCE *obs, double smooth_threshold, int scale);

extern HMM *baum_welch(HMM *new_model, HMM *old_model, 
		       SEQUENCE **obs, int n_sequence, double smooth_threshold,
		       int bakis_model, int scale);

extern HMM *bw_A(HMM *new_model, HMM *old_model, 
	  SEQUENCE **obs, int n_sequence, FB **ab,
	  double *denominator, double *inv_p);

extern HMM *bw_B(HMM *new_model, HMM *old_model, 
		 SEQUENCE **obs, int n_sequence, FB **ab,
		 double *denominator, double *inv_p);

extern HMM *bw_Pi(HMM *new_model, FB **ab, int flag);

extern HMM *bw_smooth(HMM *model, double threshold, int norm);

extern double *random_Pi(double *vector, int n);
extern double **random_A(double **matrix, int n, int m);
extern double **random_B(double **matrix, int n, int m);
extern double **random_bakis(double **matrix, int n, int m, int connections);
extern double **random_continued_bakis(double **matrix, int n, int m, 
				       int connections);

extern double probability(HMM *model, SEQUENCE **obs, 
			  int n_sequence, int scale, double *probs);

extern MULTIPLE *read_data_file(MULTIPLE *m, char *file_name);
extern SEQUENCE *read_one_sequence(FILE *fp, SEQUENCE *obs);

extern SEQUENCE *generate_sequence(HMM *model, SEQUENCE *obs, int seed);
extern void binary_search_hmm(double *vec, IntervalHMM *n, double x);
extern double *transform_vector(double *out, double *in, int n);
extern int which_one(double *vec, int n);

/*****************************************************************************/

extern double viterbi(SEQUENCE *states, HMM *model, SEQUENCE *obs);
